#!/usr/bin/perl

#Nisus Macro Block
#source front
#destination new
#text as rtf
#End Nisus Macro Block

# Created 06.11.05.
# modified 07.07.17.

use strict;
use warnings;

my $has_MacPerl;

BEGIN {
	eval {
		require MacPerl;

		import MacPerl;
	};

	# If the eval failed, we don't have the module
	if ($@) {
		$ has_MacPerl = 0;
	} else {
		$ has_MacPerl = 1;
	}
}

# global variables...:
my @analyzed_default_fonts;
my @analyzed_default_styles;

my @analyzed_this_file_fonts;
my @analyzed_this_file_styles;

my @footnote_style_flag;
my @endnote_style_flag;

my $default_ftnote_ref;
my $default_ednote_ref;

my $note_ref_style;
my $old_note_ref = "\\nisusreferencestyle";
my $new_note_ref = "\\nisusnoterefstyle";

##########

my $default_file = "$ENV{'HOME'}/Library/Application Support/Nisus Writer/Nisus New File.dot";

# my $default_file = "/Volumes/myDocument/Users/ni/Desktop/Archive/Nisus New File.dot";

open (IN, $default_file) || die ("Couldn't open Nisus New File.dot: $!.");

local undef $/;

$_ = <IN>;

close (IN);

if (/\\nisusversion3\d+/) {
	$note_ref_style = $new_note_ref;
}
else {
	$note_ref_style = $old_note_ref;
}	

my @default_fonts = &get_parts ($_, "fonts");

my $default_stylesheet = &get_parts ($_, "stylesheet");

# has the default stylesheet the footnote/endnote references...?
my ($default_ftnote_style, $default_ftnote_chr_ref) = $default_stylesheet =~ /(\\s\d+).*? \\nisusnoteplacement0.*? $note_ref_style(\d+) /s;

if ($default_ftnote_style) {
	($default_ftnote_ref) = $default_stylesheet =~ /{(\\\*\\cs$default_ftnote_chr_ref)/s;
}

my ($default_ednote_style, $default_ednote_chr_ref) = $default_stylesheet =~ /(\\s\d+).*? \\nisusnoteplacement2.*? $note_ref_style(\d+)/;

if ($default_ednote_style) {
	($default_ednote_ref) = $default_stylesheet =~ /{(\\\*\\cs$default_ednote_chr_ref)/s;
}

# analyze the default fonts...
for (@default_fonts) {
	&analyze_font ($_, 1);
}

######

# Now, load "this_file" and get font table and style sheet...

# my $this_file = shift;
# 
# my $this_file = "/Volumes/myDocument/Users/ni/Documents/Backup/Texts J/Komazawa_06-10/paper_draft.rtf";
# my $this_file = "/Volumes/myDocument/Users/ni/Desktop/Archive/TestFile.rtf";
# 
# open (IN, $this_file) || die ("Couldn't open $this_file: $!.");
# 
# $_ = <IN>;
# 
# close (IN);

$_ = <STDIN>;

if (/\\nisusversion3\d+/) {
	$note_ref_style = $new_note_ref;
}
else {
	$note_ref_style = $old_note_ref;
}	

# other global variables:
my ($no_stylesheet, $quoted_this_file_stylesheet);

# store "this file" in a temp variable...
my $temp_this_file = $_;

my $this_file_fonttbl = &get_parts ($_, "fonttbl");

my $quoted_this_file_fonttbl = quotemeta ('{\fonttbl ' . $this_file_fonttbl . "}");

my @this_file_fonts = &get_parts ($_, "fonts");
for (@this_file_fonts) {
	&analyze_font ($_, 0);
}

&merge_fonts ();

#  print $this_file_fonttbl, "\n\n", $quoted_this_file_fonttbl, "\n\n"; #exit;

my $this_file_stylesheet = &get_parts ($_, "stylesheet");

if ($this_file_stylesheet eq $default_stylesheet) { # this file doesn't have any stylesheet...
	$no_stylesheet = 1;
}
else {
	$quoted_this_file_stylesheet = quotemeta ('{\stylesheet ' . $this_file_stylesheet);
}

$this_file_stylesheet =~ s/\n?}$//s;

# has the "this file" stylesheet the footnote/endnote references...?
my ($ftnote_style, $ftnote_ref) = $this_file_stylesheet =~ /(\\s\d+).*? \\nisusnoteplacement0.*? $note_ref_style(\d+) /s;

if ($ftnote_style) {
	my ($ftnote_ref_chr_style) = $this_file_stylesheet =~ /{(\\\*\\cs$ftnote_ref)/s;
	push (@footnote_style_flag, ($ftnote_style, $ftnote_ref_chr_style));
}

my ($ednote_style, $ednote_ref) = $this_file_stylesheet =~ /(\\s\d+).*? \\nisusnoteplacement2.*? $note_ref_style(\d+)/;

if ($ednote_style) {
	my ($ednote_ref_chr_style) = $this_file_stylesheet =~ /{(\\\*\\cs$ednote_ref)/s;
	push (@endnote_style_flag, ($ednote_style, $ednote_ref_chr_style));
}

my @this_file_styles = &get_parts ($_, "");

for (@this_file_styles) {
	&analyze_style ($_, 0);
}

my @default_styles = &splitPlus (';\}[\n\r]*', $default_stylesheet);

for (@default_styles) {
	&analyze_style ($_, 1);
}

&merge_styles ();

# print $this_file_stylesheet; exit;

$temp_this_file =~ s/$quoted_this_file_fonttbl/$this_file_fonttbl/s;

if ($quoted_this_file_stylesheet) {
	$temp_this_file =~ s/$quoted_this_file_stylesheet/$this_file_stylesheet/s;
}
else { # "this_file" doesn't have a stylesheet...
	if ($temp_this_file =~ /{\\stylesheet }/) {
		$temp_this_file =~ s/{\\stylesheet }/$this_file_stylesheet/s;
	}
	else { # probably not a NWE file...
		$temp_this_file =~ /{\\colortbl.*?;}/s;
		$temp_this_file = $` . $& . $this_file_stylesheet . $';
	}
}

my $outfile = "/tmp/res_XXXXXX";
$outfile = `mktemp "$outfile"`;
$outfile =~ s/\n//g;
$outfile = $outfile . ".rtf";

open (OUT, ">$outfile");

# print $temp_this_file;

print OUT $temp_this_file;

close (OUT);

if ($has_MacPerl) { # it is later than 10.4...
	MacPerl::SetFileInfo("NISX", "RTF ", $outfile);
}

`open -a "Nisus Writer Pro" "$outfile"`;

exit;

########

sub get_parts {
	my $rtf_file = shift;
	my $return_name = shift;
	
	my ($rtfheader, $rtfdata) = $rtf_file =~ /^(.+?)({\\header.+)$/s;
	if (! $rtfheader) {
		($rtfheader, $rtfdata) = $rtf_file =~ /^(.+?)({\\footer.+)$/s;
		if (! $rtfdata) {
			($rtfheader, $rtfdata) = $rtf_file =~ /^(.+?)(\\pard.+)$/s;
		}
	}

	my ($rtf_firstheader, $fonttbl, $rtf_fonttble_footer) = $rtfheader =~ /^(\{\\rtf1.+?\{\\fonttbl ?)(\{?\\f0.+?;\}?)([\n\r]?\}[\n\r]?{\\colortbl)/s;

	if ($return_name eq "fonttbl") {
		return $fonttbl;
	}
	
	if ($return_name eq "fonts") {
		my @fonts = &splitPlus (';\}?[\n\r]*', $fonttbl);
		return @fonts;
	}
	
	my $stylesheet = &extract_rtf_token ("stylesheet", $rtfheader);
	
	$stylesheet =~ s/^{\\stylesheet //;
	
	if (! $stylesheet) { # if "this_file" doesn't have a stylesheet, use the "default_stylesheet"
		my $colortbl = &extract_rtf_token ("colortbl", $rtfheader);
		my $quoted_colortbl = quotemeta ($colortbl);
		$rtfheader =~ /$quoted_colortbl/s;
		$rtfheader = $` . $colortbl . "\n" . $default_stylesheet . $';
		$stylesheet = $default_stylesheet;
	}
	else {
		if ($stylesheet !~ /(\\s\d+|\\cs\d+)/) {
			my $quoted_stylesheet = quotemeta ($stylesheet);
			$rtfheader =~ s/$quoted_stylesheet/$default_stylesheet/s;
			$stylesheet = $default_stylesheet;
		}
	}
			
	if ($return_name eq "stylesheet") {
		return $stylesheet;
	}

	my @styles = &splitPlus (';\}[\n\r]*', $stylesheet);
	return (@styles);
}

sub analyze_font {
	my $this_font = shift;
	# \f0 \fnil \fcharset77 Times New Roman{\*\falt Times}
	my $if_default = shift;
	
	$this_font =~ s/^{//;
	$this_font =~ s/\n?;}$//s;
	my ($f_id, $charset, $f_name) = $this_font =~ /(\\f\d+) \\fnil (\\fcharset\d+) ([^{]+)/;
	if (! $f_id) {#\f0\fswiss\fcharset77 Helvetica;
		($f_id, $charset, $f_name) = $this_font =~ /(\\f\d+).*?(\\fcharset\d+) ([^{]+)/;
	}

	my %this_font = (
		id => $f_id,
		chrset => $charset,
		name => $f_name
	);
	
	if ($if_default) {
		push (@analyzed_default_fonts, {%this_font});
	}
	else {
		push (@analyzed_this_file_fonts, {%this_font});
	}
}
	

sub analyze_style {
	my $this_style = shift;
	# \s30 \nisusnoteplacement0 \nisusreferencestyle31 \sbasedon32 {\*\nsmpltxt Footnotes appear at the bottom of each page. (Click to edit this sample text.)}
# footnote text
	# \s55 \snext55 {\*\nsmpltxt Some text goes here so you can see what your style will look like.}
# \f1 \fs24 \cf1 \ql \sb0 \sa120 \sl240 \slmult1 \li0 \lin0 \fi0 \ri0 \rin0 Normal
	# \*\cs69 {\*\nsmpltxt Use this style when you want the text to appear strong. (Click to edit this sample text.)}
# \b Strong
	# \s82 \nisusnoteplacement0 \nisusreferencestyle83 
	my $if_default = shift;
	
	$this_style =~ s/^{//;
	$this_style =~ s/[\n]?;}$//s;
	
	$this_style =~ s/({\\\*\\nsmpltxt [^}]+})//s;
	my $sampletext = $1;
	$this_style =~ s/{.+?}//sg;
	$this_style =~ s/\n//gs;
	my ($s_id, $s_content, $dummy, $s_name);
	$this_style =~ s/^([\\\*c]*?s\d+) //;
	$s_id = $1;
	($s_content, $dummy, $s_name) = $this_style =~ /^((\\[A-z0-9\-]+ )+)(.+)$/;
	if (! $s_name && $this_style !~ /[{\\]+/) {
		$s_name = $this_style;
	}
	my %this_style;
	%this_style = (
		id => $s_id,
		sampletxt => $sampletext,
		content => $s_content,
		name => $s_name
	);
		
	if ($if_default) {
		push (@analyzed_default_styles, {%this_style});
	}
	else {
		push (@analyzed_this_file_styles, {%this_style});
	}
}

sub merge_fonts {
	my $this_file_font_num = $#analyzed_this_file_fonts;
	
	my %this_file_font_names;
	for my $i (0 .. $this_file_font_num) {
		my $f_name = $analyzed_this_file_fonts[$i]{name};
		my $this_file_font_id = $analyzed_this_file_fonts[$i]{id};
		$this_file_font_names{$f_name} = $this_file_font_id;
	}
	my ($old_id, $new_id);
	my $add_to_id = 1;
	for my $i (0 .. $#analyzed_default_fonts) {
		my $default_f_name = $analyzed_default_fonts[$i]{name};
		if (! exists $this_file_font_names{$default_f_name}) {
			$old_id = $analyzed_default_fonts[$i]{id};
			$old_id = quotemeta ($old_id);
			$new_id = "\\f" . ($this_file_font_num + $add_to_id);
			$add_to_id++;				
			$default_fonts[$i] =~ s/$old_id/$new_id/;
			$this_file_fonttbl .= "\n" . $default_fonts[$i];
			$default_stylesheet =~ s/$old_id/$new_id/sg; # replace all the "old_id" in 
			# default_stylesheet with the new_id...
		}
		else { # if the same font exists in both stylesheets, replace all the "old_id" in 
			# default_stylesheet with the new_id...
			$old_id = $analyzed_default_fonts[$i]{id};
			$old_id = quotemeta ($old_id);
			$new_id = $this_file_font_names{$default_f_name};
			$default_stylesheet =~ s/$old_id/$new_id/sg;
		}			
			
	}
	if ($this_file_fonttbl =~ /\n$/s) {
		$this_file_fonttbl = "{\\fonttbl " . $this_file_fonttbl . "}";
	}
	else {
		$this_file_fonttbl = "{\\fonttbl " . $this_file_fonttbl . "\n}";
	}		
}

sub merge_styles {
	my $this_file_style_num = $analyzed_this_file_styles[$#analyzed_this_file_styles]{id};
	($this_file_style_num) = $this_file_style_num =~ /(\d+)/;
	my %this_file_style_names;
	
	for my $i (0 .. $#analyzed_this_file_styles) {
		my $s_name = $analyzed_this_file_styles[$i]{name};
		$this_file_style_names{$s_name} = 1;
	}
	
	my ($old_id, $new_id);
	my $add_to_id = 1;
	my %reference_data; # for note reference...
	my $added_stylesheet;
	for my $i (0 .. $#analyzed_default_styles) {
		my $default_s_name = $analyzed_default_styles[$i]{name};
		if (! exists $this_file_style_names{$default_s_name}) {
			my $default_s_content = $analyzed_default_styles[$i]{content};
			if ((@footnote_style_flag || @endnote_style_flag) && $default_s_content =~ /$note_ref_style/) {
				next; # note reference... (note style may have arbitrary name...)
			}
			if ((@footnote_style_flag || @endnote_style_flag) && ($analyzed_default_styles[$i]{id} eq $default_ftnote_ref || $analyzed_default_styles[$i]{id} eq $default_ednote_ref)) {
				next; # note reference... (note style may have arbitrary name...)
			}
			
			my $if_chr_style = "\\s";
			$old_id = $analyzed_default_styles[$i]{id};
			my ($old_id_num) = $old_id =~ /(\d+)/;
			if ($old_id =~ /cs/) {
				$if_chr_style = '\*\cs';
			}
			$old_id = quotemeta ($old_id);
			my $new_id_num = $this_file_style_num + $add_to_id;
			$new_id = $if_chr_style . $new_id_num;
			$add_to_id++;
			$reference_data{$old_id_num} = $new_id_num;
			$default_styles[$i] =~ s/$old_id/$new_id/;
			if ($add_to_id == 2) {
				$added_stylesheet .= "\n" . $default_styles[$i];
			}
			else {
				$added_stylesheet .= $default_styles[$i];
			}
		}
	}
	
	foreach my $old_id (keys %reference_data) { # note reference...
		my $new_id = $reference_data{$old_id};
		$added_stylesheet =~ s/$note_ref_style$old_id/$note_ref_style$new_id/sg;
		$added_stylesheet =~ s/\\snext$old_id/\\snext$new_id/sg;
		$added_stylesheet =~ s/\\sbasedon$old_id/\\sbasedon$new_id/sg;
	}
	
	my $last_return = "";
	if ($added_stylesheet) {
		if ($added_stylesheet !~ /\n$/s) {
			$last_return = "\n";
		}
	}
	else { # if "this_file" has no stylesheet at all, $default_stylesheet takes the place of
		   # $this_file_stylesheet...
		if ($this_file_stylesheet !~ /\n$/s) {
			$last_return = "\n";
		}
	}
		
	if ($added_stylesheet) {
		$this_file_stylesheet = "{\\stylesheet " . $this_file_stylesheet . $added_stylesheet . $last_return . "}";
	}
	else {
		$this_file_stylesheet = "{\\stylesheet " . $this_file_stylesheet . $last_return . "}";
	}		
}
	
sub splitPlus {
	my $regex = shift;
	my $str = shift;
	
	my @res;
	while ($str) {
		my $suc = $str =~ s/^(.*?$regex)//s;
		last if (! $suc); #{
#			push (@res, $str);
# 			$str = "";
# 			last;
# 		}
		push (@res, $1);
	}
	
	return (@res);
}
		
sub extract_rtf_token {
	my $keyword = shift;
	my $str = shift;
	
	if (! $str) {
		return ("");
	}

	my $res;

	my $num = $str =~ s/^(.*?)(\{\\$keyword.+)$/$2/s;
	if (! $num || ! $str) {
#		die ("sub extract_rtf_token couldn't find the keyword...");
		return ("");
	}

	my $pat = '.*?\}';
	($res) = $str =~/^($pat)/s;
	if (! $res) {
		die ("sub extract_rtf_token couldn't find the matching close brace...");
	}
	my $temp = $res;
	my $temp2 = $temp;
	if ($temp =~ s/\{//sg == $temp2 =~ s/\}//sg) {
		return $res;
	}
	else {
		$num = 0;
		while (1) {
			$num++;
			my $pat2 = $pat x $num;
			($res) = $str =~/^(\{$pat2)/s;
			if (! $res) {
				die ("sub extract_rtf_token couldn't find the matching close brace...");
			}
			$temp = $res;
			$temp2 = $temp;
			if ($temp =~ s/\{//sg == $temp2 =~ s/\}//sg) {
				return $res;
			}
		}
	}
}

